from typing import Type

from .aido import AIDORNA
from .dnabert import DNABERT2
from .ernierna import ERNIERNA
from .hyenadna import HyenaDNA
from .nucleotide_transformer import NucleotideTransformer
from .orthrus import Orthrus
from .rinalmo import RiNALMo
from .rnabert import RNABERT
from .rnaernie import RNAErnie
from .rnafm import RNAFM
from .rnamsm import RNAMSM
from .splicebert import SpliceBERT
from .utrbert import UTRBERT
from .utrlm import UTRLM

from .embedding_model import EmbeddingModel


MODEL_CATALOG: dict[str, Type[EmbeddingModel]] = {
    "AIDO.RNA": AIDORNA,
    "DNABERT2": DNABERT2,
    "ERNIE-RNA": ERNIERNA,
    "HyenaDNA": HyenaDNA,
    "NucleotideTransformer": NucleotideTransformer,
    "RiNALMo": RiNALMo,
    "Orthrus": Orthrus,
    "RNABERT": RNABERT,
    "RNAErnie": RNAErnie,
    "RNA-FM": RNAFM,
    "RNA-MSM": RNAMSM,
    "SpliceBERT": SpliceBERT,
    "3UTRBERT": UTRBERT,
    "UTR-LM": UTRLM,
}


MODEL_VERSION_MAP: dict[str, list[str]] = {
    "AIDO.RNA": [
        "aido_rna_650m",
        "aido_rna_650m_cds",
        "aido_rna_1b600m",
        "aido_rna_1b600m_cds"
    ],
    "DNABERT2": ["dnabert2"],
    "ERNIE-RNA": ["ernierna", "ernierna-ss"],
    "HyenaDNA": [
        "hyenadna-large-1m-seqlen-hf",
        "hyenadna-medium-450k-seqlen-hf",
        "hyenadna-medium-160k-seqlen-hf",
        "hyenadna-small-32k-seqlen-hf",
        "hyenadna-tiny-16k-seqlen-d128-hf"
    ],
    "NucleotideTransformer": [
        "2.5b-multi-species",
        "2.5b-1000g",
        "500m-human-ref",
        "500m-1000g",
        "v2-50m-multi-species",
        "v2-100m-multi-species",
        "v2-250m-multi-species",
        "v2-500m-multi-species"
    ],
    "Orthrus": [
        "orthrus-large-6-track",
        "orthrus-base-4-track"
    ],
    "RiNALMo": ["rinalmo"],
    "RNABERT": ["rnabert"],
    "RNAErnie": ["rnaernie"],
    "RNA-FM": ["rna-fm", "mrna-fm"],
    "RNA-MSM": ["rnamsm"],
    "SpliceBERT": [
        "SpliceBERT.1024nt",
        "SpliceBERT-human.510nt",
        "SpliceBERT.510nt"
    ],
    "3UTRBERT": [
        "utrbert-3mer",
        "utrbert-4mer",
        "utrbert-5mer",
        "utrbert-6mer",
        "utrbert-3mer-utronly",
        "utrbert-4mer-utronly",
        "utrbert-5mer-utronly",
        "utrbert-6mer-utronly"
    ],
    "UTR-LM": [
        "utrlm-te_el",
        "utrlm-mrl",
        "utrlm-te_el-utronly",
        "utrlm-mrl-utronly"
    ]
}



from collections.abc import Callable

import numpy as np
import torch

from multimolecule import RnaTokenizer, UtrLmModel

from mrna_bench import get_model_weights_path
from mrna_bench.models import EmbeddingModel


class UTRLM(EmbeddingModel):
    """Inference wrapper for UTR-LM.

    UTR-LM is a transformer-based mRNA foundation model that is pre-trained on
    random and endogenous 5'UTR sequences from various species using MLM.

    Link: https://github.com/a96123155/UTR-LM

    This wrapper uses the multimoleule implementation of UTR-LM:
    https://multimolecule.danling.org/models/utrlm/

    It is unclear from the manuscript what the max token input is, so the value
    from multimolecule's version is used (accounting for cls/sep tokens).

    This model also offers the ability to predict solely on the 5'UTR,
    which is in-distribution for the model training data. This can be specified
    by adding '-utr' to the end of the model version.
    """

    max_length = 1026

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version.replace("_", "-")

    def __init__(self, model_version: str, device: torch.device):
        """Initialize UTR-LM inference wrapper.

        Args:
            model_version: Version of model to load. Valid versions: {
                "utrlm-te_el",
                "utrlm-mrl",
                "utrlm-te_el-utronly",
                "utrlm-mrl-utronly"
            }
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        if "-utronly" in model_version:
            self.is_sixtrack = True
        else:
            self.is_sixtrack = False

        self.tokenizer = RnaTokenizer.from_pretrained(
            "multimolecule/{}".format(model_version.replace("-utronly", "")),
            cache_dir=get_model_weights_path()
        )

        self.model = UtrLmModel.from_pretrained(
            "multimolecule/{}".format(model_version.replace("-utronly", "")),
            cache_dir=get_model_weights_path()
        ).to(device)

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using UTR-LM.

        Args:
            sequence: Sequence to be embedded.
            overlap: Nucleotide overlap between sequence chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            UTR-LM embedding of sequence with shape (1 x 128).
        """
        sequence = sequence.replace("T", "U")
        chunks = self.chunk_sequence(sequence, self.max_length - 2, overlap)

        embedding_chunks = []

        for chunk in chunks:
            toks = self.tokenizer(chunk, return_tensors="pt").to(self.device)

            chunk_output = self.model(**toks).last_hidden_state

            embedding_chunks.append(chunk_output)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(
        self,
        sequence: str,
        cds: np.ndarray,
        splice: np.ndarray,
        overlap: int = 0,
        agg_fn: Callable = torch.mean,
    ) -> torch.Tensor:
        """Embed ONLY the 5'UTR of a sequence using UTR-LM.

        Args:
            sequence: Sequence to embed.
            cds: CDS track for sequence to embed.
            splice: Unused.
            overlap: Number of overlapping nucleotides between sequence chunks.
            agg_fn: Currently unused.

        Returns:
            UTR-LM representation of 5'UTR of the sequence.
        """
        _ = splice
        five_utr_seq = self.get_fiveprime_utr(sequence, cds)

        return self.embed_sequence(five_utr_seq, overlap, agg_fn)

    def get_fiveprime_utr(self, sequence: str, cds: np.ndarray) -> str:
        """Return the portion of a sequence corresponding to the 5'UTR.

        If no CDS is detected or entire sequence is CDS, return orignal input.

        Args:
            sequence: Sequence to process.
            cds: Binary array denoting CDS.

        Returns:
            Sequence's 5'UTR, or original sequence if UTR cannot be found.
        """
        if sum(cds) == 0 or len(sequence[:np.argmax(cds)]) == 0:
            return sequence
        return sequence[:np.argmax(cds)]



from collections.abc import Callable

import torch

from multimolecule import RnaTokenizer, RiNALMoModel

from mrna_bench import get_model_weights_path
from mrna_bench.models import EmbeddingModel


class RiNALMo(EmbeddingModel):
    """Inference wrapper for RiNALMo.

    RiNALMo is a transformer-based RNA foundation model trained on 36M ncRNA
    sequences using MLM and other modern architectural improvements such as
    RoPE, SwiGLU activations, and Flash Attention.

    Link: https://github.com/lbcb-sci/RiNALMo

    This wrapper uses the multimoleule implementation of RiNALMo:
    https://huggingface.co/multimolecule/rinalmo
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize RiNALMo inference wrapper.

        Args:
            model_version: Version of model to load. Only "rinalmo" valid.
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        self.tokenizer = RnaTokenizer.from_pretrained(
            "multimolecule/rinalmo",
            cache_dir=get_model_weights_path()
        )

        self.model = RiNALMoModel.from_pretrained(
            "multimolecule/rinalmo",
            cache_dir=get_model_weights_path()
        ).to(device)

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using RiNALMo.

        Args:
            sequence: Sequence to be embedded.
            overlap: Unused.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            RiNALMo embedding of sequence with shape (1 x 1280).
        """
        if overlap != 0:
            raise ValueError("RiNALMo does not require sequence chunking.")

        sequence_in = sequence.replace("T", "U")
        toks = self.tokenizer(sequence_in, return_tensors="pt").to(self.device)

        hidden_output = self.model(**toks).last_hidden_state
        output = agg_fn(hidden_output, dim=1)

        return output

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not supported."""
        raise NotImplementedError("Six track not available for NT.")



from collections.abc import Callable

import numpy as np
import torch

from multimolecule import RnaTokenizer, UtrBertModel

from mrna_bench import get_model_weights_path
from mrna_bench.models import EmbeddingModel


class UTRBERT(EmbeddingModel):
    """Inference wrapper for 3UTRBERT.

    3UTRBERT is a transformer-based mRNA foundation model pretrained on the
    3'UTR regions of 100k RNA sequences from gencode using MLM. Various
    versions of 3UTRBERT are available with different k-mer sizes (3, 4, 5, 6).

    This wrapper also offers the option to only use the 3'UTR region of the
    sequence. This can be specified by adding '-utronly' to the end of the
    model version.

    Link: https://github.com/yangyn533/3UTRBERT

    This wrapper uses the multimolecule implementation of 3UTRBERT:
    https://huggingface.co/multimolecule/utrbert-3mer
    """

    max_length = 512

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize 3UTRBERT.

        Args:
            model_version: Version of model to load. Valid versions: {
                "utrbert-3mer",
                "utrbert-4mer",
                "utrbert-5mer",
                "utrbert-6mer",
            }. Add '-utronly' to end of version to only use 3'UTR region.
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        if "-utronly" in model_version:
            self.is_sixtrack = True
        else:
            self.is_sixtrack = False

        self.tokenizer = RnaTokenizer.from_pretrained(
            "multimolecule/{}".format(model_version.replace("-utronly", "")),
            cache_dir=get_model_weights_path()
        )
        self.model = UtrBertModel.from_pretrained(
            "multimolecule/{}".format(model_version.replace("-utronly", "")),
            cache_dir=get_model_weights_path()
        ).to(device)

        self.kmer_size = int(model_version.split("-")[1].replace("mer", ""))

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using 3UTRBERT.

        Args:
            sequence: Sequence to be embedded.
            overlap: Nucleotide overlap between sequence chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            3UTRBERT embedding of sequence with shape (1 x 768).
        """
        sequence = sequence.replace("T", "U")
        tokens = self.tokenizer(sequence, return_tensors="pt")
        input_ids = list(tokens["input_ids"][0][1:-1])

        chunks = self.chunk_tokens(input_ids, self.max_length - 2, overlap)

        embedding_chunks = []

        for chunk in chunks:
            chunk_ids = torch.Tensor([[1] + chunk + [2]])
            chunk_attn = torch.Tensor([[1] * (len(chunk) + 2)])

            toks = {
                "input_ids": chunk_ids.long().to(self.device),
                "attention_mask": chunk_attn.long().to(self.device)
            }

            cls_output = self.model(**toks).last_hidden_state[:, 0:1, :]
            embedding_chunks.append(cls_output)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(
        self,
        sequence: str,
        cds: np.ndarray,
        splice: np.ndarray,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using only 3'UTR region using 3UTRBERT.

        Args:
            sequence: Sequence to be embedded.
            cds: CDS indices.
            splice: Unused.
            overlap: Nucleotide overlap between sequence chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            3UTRBERT embedding of sequence with shape (1 x 768).
        """
        _ = splice  # Unused
        cds_sequence = self.get_threeprime_utr(sequence, cds)
        return self.embed_sequence(cds_sequence, overlap, agg_fn)

    def get_threeprime_utr(self, sequence: str, cds: np.ndarray) -> str:
        """Get 3'UTR region of sequence.

        Args:
            sequence: Sequence to extract 3'UTR from.
            cds: CDS indices.

        Returns:
            3'UTR region of sequence, or full sequence if no CDS found.
        """
        if sum(cds) == 0:
            return sequence

        cds_end = np.where(cds == 1)[0][-1] + 3

        if len(sequence[cds_end:]) < self.kmer_size:
            return sequence

        return sequence[cds_end:]



from collections.abc import Callable

import os
from pathlib import Path
import tarfile

import torch
from transformers import AutoTokenizer, AutoModel

from mrna_bench.models.embedding_model import EmbeddingModel
from mrna_bench.utils import get_data_path, download_file


MODEL_WEIGHT_URL = "https://zenodo.org/records/7995778/files/models.tar.gz"


class SpliceBERT(EmbeddingModel):
    """Inference Wrapper for SpliceBERT.

    SpliceBERT is a transformer-based RNA foundation model trained on 2 million
    vertebrate mRNA sequences using a MLM pretraining objective. Alternative
    versions are trained on only human RNA, and using smaller context windows.

    SpliceBERT 510nt versions strictly use 510nt windows, sequence that is not
    divisible by 510 is truncated.

    Link: https://github.com/biomed-AI/SpliceBERT
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        short_name_map = {
            "SpliceBERT.1024nt": "splicebert-v-1024nt",
            "SpliceBERT-human.510nt": "splicebert-v-510nt",
            "SpliceBERT.510nt": "splicebert-h-510nt"
        }
        return short_name_map[model_version]

    def __init__(self, model_version: str, device: torch.device):
        """Initialize SpliceBERT Model.

        Args:
            model_version: Model version to use. Valid versions: {
                "SpliceBERT.1024nt",
                "SpliceBERT-human.510nt",
                "SpliceBERT.510nt"
            }
            device: PyTorch device used by model inference.
        """
        super().__init__(model_version, device)

        self.is_sixtrack = False

        # Download all model weights
        weight_path = Path(get_data_path()) / "model_weights" / "splice-bert"
        weight_path.mkdir(parents=True, exist_ok=True)

        if not (weight_path / "models").exists():
            print("Fetching SpliceBERT weights.")
            dl_path, _ = download_file(MODEL_WEIGHT_URL, str(weight_path))

            with tarfile.open(dl_path) as f:
                f.extractall(weight_path)

            os.remove(dl_path)

        model_path = weight_path / "models" / model_version
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.model = AutoModel.from_pretrained(model_path).to(device)

        self.max_length = int(model_version.split(".")[1].replace("nt", ""))

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using SpliceBERT.

        SpliceBERT's 510nt models only work with a 510nt sequence length input.
        Sequences that are not divisible by 510 will have an overlap between
        the last and second last chunk.

        Args:
            sequence: Sequence to embed.
            overlap: Number of overlapping nucleotides between chunks.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            SpliceBERT representation of sequence with shape (1 x 512).
        """
        chunks = self.chunk_sequence(sequence, self.max_length, overlap)

        # Pad last chunk for 510 nt models
        if self.max_length == 510:
            if len(chunks) == 1 and len(chunks[0]) != 510:
                print(
                    "Warning: SpliceBERT-510nt input must be at least 510nts."
                    "Embedding may not work correctly."
                )
            elif len(chunks[-1]) != 510:
                overlap = 510 - len(chunks[-1])
                chunks[-1] = chunks[-2][-overlap:] + chunks[-1]
                assert len(chunks[-1]) == 510

        embedding_chunks = []

        for chunk in chunks:
            chunk_in = " ".join(list(chunk))
            input_ids = self.tokenizer.encode(chunk_in)
            input_ids = torch.as_tensor(input_ids)
            input_ids = input_ids.unsqueeze(0).to(self.device)

            last_hidden_state = self.model(input_ids).last_hidden_state

            embedding_chunks.append(last_hidden_state)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not supported."""
        raise NotImplementedError("Six track not available for SpliceBERT.")



from typing import Callable


import torch

from multimolecule import RnaErnieModel, RnaTokenizer

from mrna_bench.models import EmbeddingModel


class RNAErnie(EmbeddingModel):
    """Inference Wrapper for RNAErnie.

    RNAErnie is a transformer based RNA foundation model pre-trained using
    MLM 23M ncRNA sequences. RNAErnie uses motif-level masking during its
    pre-training by masking contiguous token regions of several sizes.

    Link: https://github.com/CatIIIIIIII/RNAErnie

    This wrapper uses the RNAErnie implementation from the multimolecule:
    https://huggingface.co/multimolecule/rnaernie
    """

    max_length = 512

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize RNAErnie inference wrapper.

        Args:
            model_version: Must be "rnaernie".
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        self.tokenizer = RnaTokenizer.from_pretrained(
            "multimolecule/{}".format(model_version)
        )

        self.model = RnaErnieModel.from_pretrained(
            "multimolecule/{}".format(model_version)
        ).to(device)

        self.is_sixtrack = False

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed RNA sequence using RNAErnie.

        Args:
            sequence: Sequence to be embedded.
            overlap: Nucleotide overlap between sequence chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            RNAErnie embedding of sequence with shape (1 x 768).
        """
        sequence = sequence.replace("T", "U")
        chunks = self.chunk_sequence(sequence, self.max_length - 2, overlap)

        embedding_chunks = []

        for chunk in chunks:
            toks = self.tokenizer(chunk, return_tensors="pt").to(self.device)

            cls_output = self.model(**toks).last_hidden_state
            embedding_chunks.append(cls_output)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not implemented for RNAErnie."""
        raise NotImplementedError("RNAErnie does not support sixtrack mode.")
